"""
Interacts for Algebra and Number Theory

AUTHORS:

- Harald Schilly (2011-01-16): initial version (:issue:`9623`)
  partially based on work by Lauri Ruotsalainen
"""

# ****************************************************************************
#       Copyright (C) 2011 Harald Schilly <harald.schilly@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from .library import polar_prime_spiral
