r"""
Concrete classes related to modules with a distinguished basis

This module provides concrete classes for various constructions
related to modules with a distinguished basis:

- :mod:`morphism` -- Concrete classes for morphisms of modules with basis

.. SEEALSO:: The category :class:`ModulesWithBasis`
"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)
del install_doc
