"""
Families

This is a backward compatibility stub. Use :mod:`sage.sets.family` instead.
"""

# Backward compatibility pointer
# Needed for unpickling.
from sage.sets.family import (Family, FiniteFamily, LazyFamily,
                              FiniteFamilyWithHiddenKeys)
