# distutils: libraries = flint
# distutils: depends = flint/acb_modular.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void psl2z_init(psl2z_t g) noexcept
    void psl2z_clear(psl2z_t g) noexcept
    void psl2z_swap(psl2z_t f, psl2z_t g) noexcept
    void psl2z_set(psl2z_t f, const psl2z_t g) noexcept
    void psl2z_one(psl2z_t g) noexcept
    bint psl2z_is_one(const psl2z_t g) noexcept
    void psl2z_print(const psl2z_t g) noexcept
    void psl2z_fprint(FILE * file, const psl2z_t g) noexcept
    bint psl2z_equal(const psl2z_t f, const psl2z_t g) noexcept
    void psl2z_mul(psl2z_t h, const psl2z_t f, const psl2z_t g) noexcept
    void psl2z_inv(psl2z_t h, const psl2z_t g) noexcept
    bint psl2z_is_correct(const psl2z_t g) noexcept
    void psl2z_randtest(psl2z_t g, flint_rand_t state, slong bits) noexcept
    void acb_modular_transform(acb_t w, const psl2z_t g, const acb_t z, slong prec) noexcept
    void acb_modular_fundamental_domain_approx_d(psl2z_t g, double x, double y, double one_minus_eps) noexcept
    void acb_modular_fundamental_domain_approx_arf(psl2z_t g, const arf_t x, const arf_t y, const arf_t one_minus_eps, slong prec) noexcept
    void acb_modular_fundamental_domain_approx(acb_t w, psl2z_t g, const acb_t z, const arf_t one_minus_eps, slong prec) noexcept
    bint acb_modular_is_in_fundamental_domain(const acb_t z, const arf_t tol, slong prec) noexcept
    void acb_modular_fill_addseq(slong * tab, slong len) noexcept
    void acb_modular_theta_transform(int * R, int * S, int * C, const psl2z_t g) noexcept
    void acb_modular_addseq_theta(slong * exponents, slong * aindex, slong * bindex, slong num) noexcept
    void acb_modular_theta_sum(acb_ptr theta1, acb_ptr theta2, acb_ptr theta3, acb_ptr theta4, const acb_t w, int w_is_unit, const acb_t q, slong len, slong prec) noexcept
    void acb_modular_theta_const_sum_basecase(acb_t theta2, acb_t theta3, acb_t theta4, const acb_t q, slong N, slong prec) noexcept
    void acb_modular_theta_const_sum_rs(acb_t theta2, acb_t theta3, acb_t theta4, const acb_t q, slong N, slong prec) noexcept
    void acb_modular_theta_const_sum(acb_t theta2, acb_t theta3, acb_t theta4, const acb_t q, slong prec) noexcept
    void acb_modular_theta_notransform(acb_t theta1, acb_t theta2, acb_t theta3, acb_t theta4, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_modular_theta(acb_t theta1, acb_t theta2, acb_t theta3, acb_t theta4, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_modular_theta_jet_notransform(acb_ptr theta1, acb_ptr theta2, acb_ptr theta3, acb_ptr theta4, const acb_t z, const acb_t tau, slong len, slong prec) noexcept
    void acb_modular_theta_jet(acb_ptr theta1, acb_ptr theta2, acb_ptr theta3, acb_ptr theta4, const acb_t z, const acb_t tau, slong len, slong prec) noexcept
    void _acb_modular_theta_series(acb_ptr theta1, acb_ptr theta2, acb_ptr theta3, acb_ptr theta4, acb_srcptr z, slong zlen, const acb_t tau, slong len, slong prec) noexcept
    void acb_modular_theta_series(acb_poly_t theta1, acb_poly_t theta2, acb_poly_t theta3, acb_poly_t theta4, const acb_poly_t z, const acb_t tau, slong len, slong prec) noexcept
    void acb_modular_addseq_eta(slong * exponents, slong * aindex, slong * bindex, slong num) noexcept
    void acb_modular_eta_sum(acb_t eta, const acb_t q, slong prec) noexcept
    int acb_modular_epsilon_arg(const psl2z_t g) noexcept
    void acb_modular_eta(acb_t r, const acb_t tau, slong prec) noexcept
    void acb_modular_j(acb_t r, const acb_t tau, slong prec) noexcept
    void acb_modular_lambda(acb_t r, const acb_t tau, slong prec) noexcept
    void acb_modular_delta(acb_t r, const acb_t tau, slong prec) noexcept
    void acb_modular_eisenstein(acb_ptr r, const acb_t tau, slong len, slong prec) noexcept
    void acb_modular_elliptic_k(acb_t w, const acb_t m, slong prec) noexcept
    void acb_modular_elliptic_k_cpx(acb_ptr w, const acb_t m, slong len, slong prec) noexcept
    void acb_modular_elliptic_e(acb_t w, const acb_t m, slong prec) noexcept
    void acb_modular_elliptic_p(acb_t wp, const acb_t z, const acb_t tau, slong prec) noexcept
    void acb_modular_elliptic_p_zpx(acb_ptr wp, const acb_t z, const acb_t tau, slong len, slong prec) noexcept
    void acb_modular_hilbert_class_poly(fmpz_poly_t res, slong D) noexcept
