# distutils: libraries = flint
# distutils: depends = flint/nmod_poly_factor.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void nmod_poly_factor_init(nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_clear(nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_realloc(nmod_poly_factor_t fac, slong alloc) noexcept
    void nmod_poly_factor_fit_length(nmod_poly_factor_t fac, slong len) noexcept
    void nmod_poly_factor_set(nmod_poly_factor_t res, const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_print(const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_insert(nmod_poly_factor_t fac, const nmod_poly_t poly, slong exp) noexcept
    void nmod_poly_factor_concat(nmod_poly_factor_t res, const nmod_poly_factor_t fac) noexcept
    void nmod_poly_factor_pow(nmod_poly_factor_t fac, slong exp) noexcept
    bint nmod_poly_is_irreducible(const nmod_poly_t f) noexcept
    bint nmod_poly_is_irreducible_ddf(const nmod_poly_t f) noexcept
    bint nmod_poly_is_irreducible_rabin(const nmod_poly_t f) noexcept
    bint _nmod_poly_is_squarefree(mp_srcptr f, slong len, nmod_t mod) noexcept
    bint nmod_poly_is_squarefree(const nmod_poly_t f) noexcept
    void nmod_poly_factor_squarefree(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    bint nmod_poly_factor_equal_deg_prob(nmod_poly_t factor, flint_rand_t state, const nmod_poly_t pol, slong d) noexcept
    void nmod_poly_factor_equal_deg(nmod_poly_factor_t factors, const nmod_poly_t pol, slong d) noexcept
    void nmod_poly_factor_distinct_deg(nmod_poly_factor_t res, const nmod_poly_t poly, slong * const * degs) noexcept
    void nmod_poly_factor_distinct_deg_threaded(nmod_poly_factor_t res, const nmod_poly_t poly, slong * const * degs) noexcept
    void nmod_poly_factor_cantor_zassenhaus(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void nmod_poly_factor_berlekamp(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void nmod_poly_factor_kaltofen_shoup(nmod_poly_factor_t res, const nmod_poly_t poly) noexcept
    mp_limb_t nmod_poly_factor_with_berlekamp(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor_with_cantor_zassenhaus(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor_with_kaltofen_shoup(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    mp_limb_t nmod_poly_factor(nmod_poly_factor_t res, const nmod_poly_t f) noexcept
    void _nmod_poly_interval_poly_worker(void * arg_ptr) noexcept
