from collections.abc import Sequence
from typing import Any

class SymbolicSeries:
    def __init__(self, SR: Any) -> None: ...
    def is_terminating_series(self) -> bool: ...
    def truncate(self) -> Any: ...
    def default_variable(self) -> Any: ...
    def coefficients(self, x: Any = None, sparse: bool = True) -> list[list[Any, int]] | list[Any]: ...
    def power_series(self, base_ring: Any) -> Any: ...
