from collections.abc import Iterable
from typing import Any

def lcm(a: Any, b: Any = None) -> Any: ...
def LCM_list(v: Iterable[Any]) -> Any: ...
def LCM_generic(itr: Iterable[Any], ret: Any) -> Any: ...
