# distutils: libraries = flint
# distutils: depends = flint/bernoulli.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void bernoulli_rev_init(bernoulli_rev_t it, ulong n) noexcept
    void bernoulli_rev_next(fmpz_t numer, fmpz_t denom, bernoulli_rev_t it) noexcept
    void bernoulli_rev_clear(bernoulli_rev_t it) noexcept
    void bernoulli_fmpq_vec_no_cache(fmpq * res, ulong a, slong num) noexcept
    void bernoulli_cache_compute(slong n) noexcept
    slong bernoulli_bound_2exp_si(ulong n) noexcept
    ulong bernoulli_mod_p_harvey(ulong n, ulong p) noexcept
    void _bernoulli_fmpq_ui_zeta(fmpz_t num, fmpz_t den, ulong n) noexcept
    void _bernoulli_fmpq_ui_multi_mod(fmpz_t num, fmpz_t den, ulong n, double alpha) noexcept
    void _bernoulli_fmpq_ui(fmpz_t num, fmpz_t den, ulong n) noexcept
    void bernoulli_fmpq_ui(fmpq_t b, ulong n) noexcept
