#########################################################################
#       Copyright (C) 2004--2006 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  http://www.gnu.org/licenses/
#########################################################################

# The default precision for computation and display of q-expansions of
# modular forms.
from sage.rings.integer import Integer
DEFAULT_PRECISION = Integer(6)
DEFAULT_VARIABLE = 'q'
