from typing import Any

class PynacConstant:
    pointer: Any
    _object: Any
    _name: str

    def __cinit__(self, name: str, texname: str, domain_string: str) -> None: ...
    def __dealloc__(self) -> None: ...
    def serial(self) -> int: ...
    def name(self) -> str: ...
    def __repr__(self) -> str: ...
    def expression(self) -> Any: ...
