"""
Interacts for Calculus

AUTHORS:

- Harald Schilly (2011-01-16): initial version (#9623) partially based on work by Lauri Ruotsalainen
"""

# ****************************************************************************
#       Copyright (C) 2011 Harald Schilly <harald.schilly@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .library import taylor_polynomial
from .library import definite_integral
from .library import function_derivative, difference_quotient, quadratic_equation
from .library import trigonometric_properties_triangle
from .library import secant_method, newton_method, trapezoid_integration, simpson_integration, bisection_method
from .library import riemann_sum
from .library import function_tool
