# distutils: libraries = flint
# distutils: depends = flint/mpfr_vec.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    mpfr_ptr _mpfr_vec_init(slong len, flint_bitcnt_t prec) noexcept
    void _mpfr_vec_clear(mpfr_ptr vec, slong len) noexcept
    void _mpfr_vec_zero(mpfr_ptr vec, slong len) noexcept
    void _mpfr_vec_set(mpfr_ptr vec1, mpfr_srcptr vec2, slong len) noexcept
    void _mpfr_vec_add(mpfr_ptr res, mpfr_srcptr vec1, mpfr_srcptr vec2, slong len) noexcept
    void _mpfr_vec_scalar_mul_mpfr(mpfr_ptr res, mpfr_srcptr vec, slong len, mpfr_t c) noexcept
    void _mpfr_vec_scalar_mul_2exp(mpfr_ptr res, mpfr_srcptr vec, slong len, flint_bitcnt_t exp) noexcept
    void _mpfr_vec_scalar_product(mpfr_t res, mpfr_srcptr vec1, mpfr_srcptr vec2, slong len) noexcept
